/* 
 * Copyright (C) 2004 Romain Dolbeau <romain@dolbeau.org>
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

package org.dolbeau.mencgen;

import java.util.*;
import java.io.*;
import javax.swing.*;

import org.dolbeau.xsCOEj.JobDist;
import org.dolbeau.xsCOEj.Option;
import org.dolbeau.xsCOEj.OptionFactoryXML;
import org.dolbeau.xsCOEj.OptionInstance;
import org.dolbeau.xsCOEj.OptionInstanceList;
import org.dolbeau.xsCOEj.OptionInstanceListBreeder1;
import org.dolbeau.xsCOEj.ResultsPool;
import org.dolbeau.xsCOEj.StatisticsDisplayPanel;

import org.w3c.dom.Document;

public class lavcGeneticMain {

    /* number of tests to run initially. each subsequent path run 1/3 that number */
    /* can be changed in UI */
    public static final int kSimultaneousTests = 3 * 5; // must be multiple of 3
    /* source file & non-encoding options */
    /* can be changed in UI */
    
    public static final String kDefaultSource =
        "/passauvegarde/kili/dolbeau/DVDs/SIN_CITY_LOCATION";
    public static final String kDefaultExtraOpts = 
        "-frames 100 -ss 1:00:00 -vf crop=704:544:10:18,scale=640:352 dvd://5 -dvd-device ";
    
    /* maximum number of pass (maximum number of run will be (kMaxPass + 2) * (kSimultaneousTests / 3) */
    /* only a safety, as the UI can stop the run */
    public static final int kMaxPass = 1000000;
    /* percent chance of randomizing each option */
    /* can be changed in UI */
    public static final int kRandomBreeding = 10;
    /* percentage within best PSNR to look for faster result */
    /* can be changed in UI */
    public static final double kWithinPercentage = 1.0;
    /* max number of attempts when randomizing (in case the generated combination is already done) */
    public static final int kMaxAttempts = 8;
    /* filename to take the machine list (one hostname per line) */
    public static final String kMachineFilename = "machine.list";
    /* default bitrate for encoding */
    /* can be changed in UI */
    public static final int kDefaultBitrate = 1250;
    /* default workdir */
    /* can be changed in UI */
    public static final String kDefaultWorkdir = "/tmp";
    
    // private static Set allCombinations = new HashSet();
    private static Random random = new Random(0); // known seed for now, so results are reproductible

    // shared by all
    private static JobDist ljd;

    static void err_println(String s) {
        synchronized (System.err) {
            System.err.println(s);
        }
    }

    static void out_println(String s) {
        synchronized (System.out) {
            System.out.println(s);
        }
    }

    private static List getMachineList(String filename) {
        List theSet = new ArrayList();
        try {
             InputStream is = new FileInputStream(filename);
             BufferedReader br = new BufferedReader(new InputStreamReader(is));
             String line = br.readLine();
             while (line != null) {
                 theSet.add(line);
                 line = br.readLine();
             }
             br.close();
             is.close();
        } catch (Exception e) {
            lavcGeneticMain.err_println("Exception " + e + " while reading machines file");
            theSet.clear();
            theSet.add("localhost");
        }
        
        return theSet;
    }

    private static void computeThemAll(Set newList, JobDist ljd, ResultsPool lrp, lavcCodecData lcd, int bitrate, boolean twopass, String workdir, String sourceFile, String extraOpts) {
        int nSimultaneousTests = newList.size();
        Thread[] allThreads = new Thread[nSimultaneousTests];
        Iterator it = newList.iterator();
        int j = 0;
        while (it.hasNext()) {
            OptionInstanceList options = (OptionInstanceList)it.next();

            //lavcGeneticMain.err_println("TRYING (" + j + "): " + options);
            
            lavcEncodeData led = new lavcEncodeData(options, lcd);
            led.sourceFile = sourceFile;
            led.extraOpts = extraOpts;
            led.workdir = workdir + "/" + led.getLacvOptsString().replace(':', '/');
            led.bitrate = bitrate;
            
            allThreads[j] = new lavcTestThread(j, led, ljd, lrp, twopass);
            
            allThreads[j].start();
            j++;
        }
        for (int i = 0; i < j ; i++) {
            try {
                //lavcGeneticMain.err_println("WAITING FOR: " + i);
                allThreads[i].join();
            } catch (Exception e) { lavcGeneticMain.err_println("Exception " + e + " while waiting for thread #" + i); }
        }
    }

    static void loadOptsFromFile(String path, Set allOptions, Set allDefaultLists) {
        try {
            File fileReader = new File(path);

            javax.xml.parsers.DocumentBuilder db =
                javax.xml.parsers.DocumentBuilderFactory.newInstance().newDocumentBuilder();

            Document doc = db.parse(fileReader);

            OptionFactoryXML ofx = new OptionFactoryXML(doc);
      
            allOptions.addAll(ofx.getAllOptions());
            allDefaultLists.addAll(ofx.getDefaultLists());
      
        }
        catch (java.io.FileNotFoundException fnfe) { System.err.println("Open failed (FileNotFoundException): " + fnfe); }
        catch (java.io.IOException ie)  { System.err.println("Open failed (IOException): " + ie); }
        catch (javax.xml.parsers.ParserConfigurationException pce)  { System.err.println("Open failed (ParserConfigurationException): " + pce); }
        catch (org.xml.sax.SAXException se)  { System.err.println("Open failed (SAXException): " + se); }
    }

    public static void main(String[] arg) {
        List machineList = getMachineList(kMachineFilename);

        lavcGeneticMain.err_println("Machines: " + machineList);

        ljd = new JobDist(machineList);

        createNewMainDisplay();
    }

    static long numberOfCombinations(Set allSelectedOptions) {
        long allCombinationsCount = 1;
        Iterator it = allSelectedOptions.iterator();
        while (it.hasNext()) {
            Option o = (Option)it.next();
            List oAll = o.getAllInstances();
            allCombinationsCount = allCombinationsCount * oAll.size();
        }
        return allCombinationsCount;
    }
    
    static boolean done = false;
    static final Object doneO = new Object(); 
    static boolean isDone() {
        boolean temp = false;
        synchronized (doneO) {
            temp = done;
        }
        return temp;
    }
    static void setDone() {
        synchronized (doneO) {
            done = true;;
        }
    }
    
    static class allCodecsBarrier {
        private Object waitO = new Object();
        private int registered = 0;
        private int step = 0;
        private int doneInStep = 0;

        allCodecsBarrier(int r) {
            registered = r;
        }

        void wait(int s) {
            synchronized(waitO) {
                if (s > step) {
                    doneInStep = 0;
                    step = s;
                }
                if (s == step) {
                    doneInStep ++;
                    if (doneInStep >= registered) {
                        lavcGeneticMain.err_println("WAKING everyone on step " + step);
                        waitO.notifyAll();
                    }
                    else {
                        lavcGeneticMain.err_println("WAITING (" + doneInStep + ") on step " + step);
                        try {
                            waitO.wait();
                        } catch (Exception e) { }
                    }
                } else {
                    lavcGeneticMain.err_println("ERROR: bad step !\n");
                    waitO.notifyAll();
                    doneInStep = registered;
                    step = s;
                }
            }
        }
    }

    static void doTheJob(String codec, Set allSelectedOptions, StatisticsDisplayPanel lasd, lavcCodecData lcd,
                         int nSimultaneousTests, int nRandomBreeding, int bitrate, boolean twopass,
                         String workdir, double nWithinPercentage, lavcGeneticMain.allCodecsBarrier acb,
                         String sourceFile, String extraOpts) {
        ResultsPool allResults = new ResultsPool();
        Iterator it;

        long allCombinationsCount = numberOfCombinations(allSelectedOptions);
        
        lavcGeneticMain.err_println("[" + codec + "] " +"Total options combinations: " + allCombinationsCount);

        lavcGeneticMain.err_println("[" + codec + "] " +"Parameters:\n" +
                                    "\t# of tests: " + nSimultaneousTests + "\n" +
                                    "\tRandom Breeding: " + nRandomBreeding + " %\n" +
                                    "\tBitrate: " + bitrate + "\n" + 
                                    "\tTwo-Pass: " + (twopass ? "yes" : "no") + "\n" +
                                    "\tBase workdir: " + workdir + "\n" +
                                    "\tApprox. Percentage: " + nWithinPercentage);

        Thread[] allThreads = new Thread[nSimultaneousTests];

        // **** STEP ZERO *****
        int step = 0;
        Set newList = new TreeSet();


        if ((lcd.allDefaultLists == null) ||
            (lcd.allDefaultLists.size() == 0)) {
            newList.add(OptionInstanceList.buildDefaultList(allSelectedOptions));
        } else {
            it = lcd.allDefaultLists.iterator();
            while (it.hasNext()) {
                OptionInstanceList l = (OptionInstanceList)it.next();
                OptionInstanceList l2 = l.updateToNewOptionSet(allSelectedOptions);
                if (l2 != null)
                    newList.add(l2);
            }
        }
        
        System.err.println("Using default(s) lists: " + newList);
        
        computeThemAll(newList, ljd, allResults, lcd, bitrate, twopass, workdir, sourceFile, extraOpts);

        lavcGeneticMain.err_println("[" + codec + "] " +"Results computed so far: " + allResults.numberOfResults() +
                                    " (" + (100. * (double)allResults.numberOfResults() / (double)allCombinationsCount) + " % of all)");
        OptionInstanceList br = allResults.getBestResultKey();
        lavcGeneticMain.err_println("[" + codec + "] " +"Best Result (" + step + "): " + br + " [" + allResults.getResult(br) + "]");
        OptionInstanceList fwpb = allResults.fasterWithinPercentBestKey(nWithinPercentage);
        lavcGeneticMain.err_println("[" + codec + "] " +"Faster within " + nWithinPercentage + "% (" + step + "): " +
                                    fwpb + " [" + allResults.getResult(fwpb) + "]");
        lavcGeneticMain.err_println("[" + codec + "] " +"Statistics:\n" + allResults.optionStatistics(allSelectedOptions));

        if (lasd != null) {
            lasd.refresh(allResults, br, step, step, fwpb, nWithinPercentage, allResults);
        }

        acb.wait(step);

        // **** STEP ONE ***** 

        step++; // 1

        newList.clear();

        for (int i = allResults.numberOfResults(); i < nSimultaneousTests; i++) {
            OptionInstanceList lol = null;
            int j = 0;
            do {
                lol = OptionInstanceList.buildRandomList(allSelectedOptions, random);
                j++;
            } while (((lol == null) || newList.contains(lol)) && (j < kMaxAttempts));
            if ((lol != null) && (!newList.contains(lol))) {
                newList.add(lol);
                //lavcGeneticMain.err_println("[" + codec + "] random list generated after " + j + " attempt(s)");
            } else {
                if (lol == null) {
                    //lavcGeneticMain.err_println("[" + codec + "] random list FAILURE after " + j + " attempt(s)");
                } else {
                    //lavcGeneticMain.err_println("[" + codec + "] random list CONFLICT after " + j + " attempt(s)");
                }
            }
        }

        computeThemAll(newList, ljd, allResults, lcd, bitrate, twopass, workdir, sourceFile, extraOpts);
        
        lavcGeneticMain.err_println("[" + codec + "] " +"Results computed so far: " + allResults.numberOfResults() +
                                    " (" + (100. * (double)allResults.numberOfResults() / (double)allCombinationsCount) + " % of all)");
        br = allResults.getBestResultKey();
        lavcGeneticMain.err_println("[" + codec + "] " +"Best Result (" + step + "): " + br + " [" + allResults.getResult(br) + "]");
        fwpb = allResults.fasterWithinPercentBestKey(nWithinPercentage);
        lavcGeneticMain.err_println("[" + codec + "] " +"Faster within " + nWithinPercentage + "% (" + step + "): " +
                                    fwpb + " [" + allResults.getResult(fwpb) + "]");
        lavcGeneticMain.err_println("[" + codec + "] " +"Statistics:\n" + allResults.optionStatistics(allSelectedOptions));

        if (lasd != null) {
            lasd.refresh(allResults, br, step, step, fwpb, nWithinPercentage, allResults);
        }

        acb.wait(step);

        // ***** STEP TWO AND MORE ***** 

        int bestStep = step;
        ResultsPool allButN = new ResultsPool(allResults);

        while (!isDone()) {
            step ++;

            int removeNumber = (nSimultaneousTests * 2 / 3) - (nSimultaneousTests - allButN.numberOfResults());

            allButN = allButN.getAllButHalfTheWorstN(removeNumber, random);

            //lavcGeneticMain.err_println("[" + codec + "] " +"ALL BUT N (" + validResult * 2 / 3 + "): " + allButN);

            int keepUpFactor = (nSimultaneousTests * 2 / 3) - allButN.numberOfResults();
            
            newList = OptionInstanceListBreeder1.breedThemAll(allButN.getSetOfKeys(), (nSimultaneousTests / 3) + keepUpFactor, allResults, nRandomBreeding, allSelectedOptions, random);

            if (newList.size() == 0) { // we failed generating anything, try fully random
                newList = OptionInstanceListBreeder1.breedThemAll(allButN.getSetOfKeys(), (nSimultaneousTests / 3) + keepUpFactor, allResults, 100, allSelectedOptions, random);
            }

            lavcGeneticMain.err_println("[" + codec + "] " +"Step " + step + " starts " + newList.size() + " runs (" +  (nSimultaneousTests / 3) + " + " + keepUpFactor + " asked)");
            
            if ((newList.size() == 0) || (step >= kMaxPass))
                setDone();
            
            //lavcGeneticMain.err_println("[" + codec + "] " +"CHILDREN: " + newList);
            if (newList.size() > 0) {
                computeThemAll(newList, ljd, allButN, lcd, bitrate, twopass, workdir, sourceFile, extraOpts);
                allResults.mergeResults(allButN);
            
                //lavcGeneticMain.err_println("[" + codec + "] " +"ALL Results (" + step + "): " + allResults.getAllResults());
                OptionInstanceList br2 = allButN.getBestResultKey();
                boolean unchanged = true;
                if (br2 != null)
                    unchanged = br2.equals(br);
                else
                    setDone();
                if (!unchanged)
                    bestStep = step;
                lavcGeneticMain.err_println("[" + codec + "] " +"Results computed so far: " + allResults.numberOfResults() + " (" + (100. * (double)allResults.numberOfResults() / (double)allCombinationsCount) + " % of all)");
                lavcGeneticMain.err_println("[" + codec + "] " +"Best Result (" + step + "): " + br2 + " [" + allButN.getResult(br2) + "]" + 
                                            (unchanged ? " {UNCHANGED since " + bestStep + "}" : " {IMPROVED}"));
                fwpb = allButN.fasterWithinPercentBestKey(nWithinPercentage);
                lavcGeneticMain.err_println("[" + codec + "] " +"Faster within " + nWithinPercentage + "% (" + step + "): " + fwpb + " [" + allButN.getResult(fwpb) + "]");
                lavcGeneticMain.err_println("[" + codec + "] " +"Statistics (best only):\n" + allButN.optionStatistics(allSelectedOptions));
                
                if (lasd != null) {
                    lasd.refresh(allButN, br, step, bestStep, fwpb, nWithinPercentage, allResults);
                }
                br = br2;
            }

            acb.wait(step);
        }
        br = allResults.getBestResultKey();
        fwpb = allResults.fasterWithinPercentBestKey(nWithinPercentage);
        if (lasd != null) {
            lasd.refresh(allResults.getNBestResults(nSimultaneousTests),
                         br, -1, bestStep, fwpb, nWithinPercentage, allResults);
        }
        try {
            File resultsFile = new File(workdir + "/" + "lavcGenetic.results." + codec + "." + new Date() + ".log");
            FileWriter fw = new FileWriter(resultsFile);
            fw.write("\tSource File: " + sourceFile + "\n");
            fw.write("\tExtra Opts: " + extraOpts + "\n");
            fw.write("\tBitrate: " + bitrate + "\n");
            fw.write("\tWorkdir: " + workdir + "\n");
            fw.write("\t2-Pass: " + (twopass ? "yes" : "no") + "\n");
            fw.write("----------------------------------------\n");
            allResults.dumpEachOptionBenefit(fw, br, allSelectedOptions);
            fw.write("----------------------------------------\n");
            allResults.dumpAllResults(fw);
            fw.close();
        } catch (IOException e) { lavcGeneticMain.err_println("Exception while trying to write results log: " + e); }
    }

    public static void createNewMainDisplay() {
        final JFrame frame = new JFrame("lavcGenetic");
        final Map codecMap = new HashMap();
        lavcCodecData lcd;

        Set lavcOptions = new TreeSet();
        Set lavcDefLists = new TreeSet();
        loadOptsFromFile("mencoder.lavc.options.xml", lavcOptions, lavcDefLists);
        Set xvidOptions = new TreeSet();
        Set xvidDefLists = new TreeSet();
        loadOptsFromFile("mencoder.xvid.options.xml", xvidOptions, xvidDefLists);
        Set x264Options = new TreeSet();
        Set x264DefLists = new TreeSet();
        loadOptsFromFile("mencoder.x264.options.xml", x264Options, x264DefLists);
        
        lcd = new lavcCodecData("lavc", "-lavcopts", lavcOptions, lavcDefLists, "vpass", "vbitrate");
        codecMap.put("libavcodec", lcd);
        lcd = new lavcCodecData("xvid", "-xvidencopts", xvidOptions, xvidDefLists, "pass", "bitrate");
        codecMap.put("XviD", lcd);
        lcd = new lavcCodecData("x264", "-x264encopts", x264Options, x264DefLists, "pass", "bitrate");
        codecMap.put("x264", lcd);
        lcd = new lavcCodecData("lavc_copy", "-lavcopts", lavcOptions, lavcDefLists, "vpass", "vbitrate");
        codecMap.put("libavcodec_copy", lcd);
        
        
        final JPanel mainDisplay = new lavcGeneticMainDisplay(frame, codecMap);

        frame.getContentPane().setLayout(new BoxLayout(frame.getContentPane(), javax.swing.BoxLayout.Y_AXIS));
        frame.getContentPane().add(mainDisplay);
        frame.pack();
        frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
    }

    /* 
   static void dumpEachOptionBenefit(Writer w,
                                      OptionInstanceList l,
                                      Set allSelectedOptions,
                                      ResultsPool lrp) throws java.io.IOException {
        lavcResults rl = (lavcResults)lrp.getResult(l);
        
        w.write(l + " := " + rl + "\n");

        Iterator it = allSelectedOptions.iterator();
        while (it.hasNext()) {
            Option opt = (Option)it.next();
            List allInstances = opt.getAllInstances();
            if (allInstances.size() > 1) {
                OptionInstance mine = l.getOptionInstance(opt);
                w.write("\tOption: " + opt.getDescription() + " (has: " + mine + ")\n");
                Iterator it2 = allInstances.iterator();
                while (it2.hasNext()) {
                    OptionInstance loi = (OptionInstance)it2.next();
                    if (loi.compareTo(mine) != 0) {
                        OptionInstanceList l2 = new OptionInstanceList(l);
                        l2.remove(mine);
                        l2.add(loi);
                        Collections.sort(l2);
                        lavcResults rl2 = (lavcResults)lrp.getResult(l2);
                        if (rl2 != null) {
                            w.write("\t\t[" + loi + "] := " + rl2 +
                                    " (PSNR All gain: " + 
                                    (rl.psnrAll - rl2.psnrAll) +
                                    ", " + (((rl.psnrAll - rl2.psnrAll) / (rl2.psnrAll)) * 100.) + " %)" +
                                    " (Time loss: " +
                                    (rl.getSeconds() - rl2.getSeconds()) +
                                    ", " + (((rl.getSeconds() - rl2.getSeconds()) / (rl2.getSeconds())) * 100.) + " %)" +
                                    "\n");
                        } else {
                            w.write("\t\t[" + loi + "] not tested\n");
                        }
                    }
                }
            }
        }
    }
    */
}
