/* 
 * Copyright (C) 2004 Romain Dolbeau <romain@dolbeau.org>
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

package org.dolbeau.mencgen;

import java.util.*;
import javax.swing.*;
import javax.swing.text.*;
import java.awt.event.*;
import java.io.*;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeEvent;

import org.dolbeau.xsCOEj.Option;
import org.dolbeau.xsCOEj.OptionChooser;
import org.dolbeau.xsCOEj.StatisticsDisplayPanel;

class lavcGeneticMainDisplay extends JPanel {
    private final JFrame parent;
    final private java.text.NumberFormat numberFormat = java.text.NumberFormat.getIntegerInstance();
    private final Box stBox = Box.createHorizontalBox();
    private final Box rbBox = Box.createHorizontalBox();

    private final Box bwBox = Box.createHorizontalBox();
    private final Box foiBox = Box.createHorizontalBox();
    private final Box bwAndfoiBox = Box.createHorizontalBox();

    private final Map codecMap;
    private final Map chooserMap = new HashMap();
    private final Set selectedCodec = new HashSet();

    private final JTabbedPane jtb = new JTabbedPane();
    private final JScrollPane mainPane = new JScrollPane(jtb);
    private final Box bBox = Box.createHorizontalBox();
    private final Box dcBox = Box.createHorizontalBox();

    private int nSimultaneousTests = lavcGeneticMain.kSimultaneousTests;
    private int nRandomBreeding = lavcGeneticMain.kRandomBreeding;
    private int bitrate = lavcGeneticMain.kDefaultBitrate;
    private boolean twopass = true;
    private String workdir = lavcGeneticMain.kDefaultWorkdir;
    private int nWithinPercentage = (int)(lavcGeneticMain.kWithinPercentage * 100.);

    private String sourceFile = lavcGeneticMain.kDefaultSource;
    private String extraOpts = lavcGeneticMain.kDefaultExtraOpts;

    private final JLabel directoryChooserLabel = new JLabel(" Workdir: " + workdir + " ");
    private final JLabel sourceFileChooserLabel = new JLabel(" Source File: " + sourceFile + " ");
    private final JTextField jtfExtraOpts = new JTextField(extraOpts);

    public static final int MIN_BITRATE = 500;
    public static final int MAX_BITRATE = 1800;
    final private JPanel btBox = new JPanel();
    final private NumberFormatter btFormatter = new NumberFormatter(numberFormat);
    {
        btFormatter.setMinimum(new Integer(MIN_BITRATE));
        btFormatter.setMaximum(new Integer(MAX_BITRATE));
    }
    final private JFormattedTextField btTextField = new JFormattedTextField(btFormatter);
    {
        btTextField.setColumns(5);
    }
    final private JSlider btSlider;
    
    class lavcSimultaneousTestsSliderListener implements javax.swing.event.ChangeListener {
        public void stateChanged(javax.swing.event.ChangeEvent e) {
            JSlider source = (JSlider)e.getSource();
            if (!source.getValueIsAdjusting()) {
                nSimultaneousTests = (int)source.getValue();
            }    
        }
    }

    class lavcRandomBreedingSliderListener implements javax.swing.event.ChangeListener {
        public void stateChanged(javax.swing.event.ChangeEvent e) {
            JSlider source = (JSlider)e.getSource();
            if (!source.getValueIsAdjusting()) {
                nRandomBreeding = (int)source.getValue();
            }
        }
    }

    class lavcBitrateSliderListener implements javax.swing.event.ChangeListener {
        public void stateChanged(javax.swing.event.ChangeEvent e) {
            JSlider source = (JSlider)e.getSource();
            if (!source.getValueIsAdjusting()) {
                bitrate = (int)source.getValue();
                btTextField.setValue(new Integer(bitrate));
            } else {
                btTextField.setText(String.valueOf(new Integer((int)source.getValue())));
            }    
        }
    }

    class lavcBestWithinSliderListener implements javax.swing.event.ChangeListener {
        public void stateChanged(javax.swing.event.ChangeEvent e) {
            JSlider source = (JSlider)e.getSource();
            if (!source.getValueIsAdjusting()) {
                nWithinPercentage = (int)source.getValue();
            }
        }
    }

    static double lavcFloatIntervalCompute(int v) {
        int power = 1 + (v / 3);
        int divby2 = v % 3;
        double value = (1. / (Math.pow(10., (double)power)));
        switch (divby2) {
        case 0:
        case 4:
            // value = 4. * value / 4.;
            break;
        case 1:
            value = 2. * value / 4.;
            break;
        case 2:
            value = 1. * value / 4.;
            break;
        default:
            break;
        }
        return value;
    }

    class lavcFloatIntervalSliderListener implements javax.swing.event.ChangeListener {
        public void stateChanged(javax.swing.event.ChangeEvent e) {
            JSlider source = (JSlider)e.getSource();
            if (!source.getValueIsAdjusting()) {
                int v = source.getValue();
                setAverageParameterAllOptions(new Double(lavcFloatIntervalCompute(v)));
            }
        }
    }

    class directoryChooserAction extends AbstractAction {
        directoryChooserAction(String n) {
            super(n);
        }

        public void actionPerformed(ActionEvent evt) {
            try {
                final JFileChooser fc = new JFileChooser(workdir);
                fc.setFileSelectionMode(JFileChooser.DIRECTORIES_ONLY);
                int returnVal = fc.showOpenDialog(parent);
                
                if (returnVal == JFileChooser.APPROVE_OPTION) {
                    File file = fc.getSelectedFile();
                    String path = file.getCanonicalPath();
                    workdir = path;
                    directoryChooserLabel.setText(" Workdir: " + path + " ");
                } else {
                }
            } catch (Exception e) { System.err.println("Error: " + e); }
        }
    }

    class sourceFileChooserAction extends AbstractAction {
        sourceFileChooserAction(String n) {
            super(n);
        }

        public void actionPerformed(ActionEvent evt) {
            try {
                final JFileChooser fc = new JFileChooser(sourceFile);
                fc.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
                int returnVal = fc.showOpenDialog(parent);
                
                if (returnVal == JFileChooser.APPROVE_OPTION) {
                    File file = fc.getSelectedFile();
                    String path = file.getCanonicalPath();
                    sourceFileChooserLabel.setText(" Source File: " + path + " ");
                    sourceFile = path;
                } else {
                }
            } catch (Exception e) { System.err.println("Error: " + e); }
        }
    }

    class codecTabbedPaneChangeListener implements javax.swing.event.ChangeListener {
        public void stateChanged(javax.swing.event.ChangeEvent e) {
            JTabbedPane source = (JTabbedPane)e.getSource();
            java.awt.Component component = source.getSelectedComponent();
        }
    }

    class selectCodecAction extends AbstractAction {
        private String name;

        selectCodecAction(String n) {
            super(n);
            name = n;
        }
        public void actionPerformed(ActionEvent evt) {
            JCheckBox cb = (JCheckBox)evt.getSource();
            if (cb.isSelected()) {
                selectedCodec.add(name);
            } else {
                selectedCodec.remove(name);
            }
        }
    }

    void addAllOptions(String codec, Box chooserBox) {
        chooserBox.removeAll();
        
        Set allChooser = (Set)chooserMap.get(codec);
        if (allChooser == null) {
            allChooser = new TreeSet();
            chooserMap.put(codec, allChooser);
            Set allOptions = ((lavcCodecData)codecMap.get(codec)).allOptions;
            List allOptions2 = new ArrayList(allOptions);
            Collections.sort(allOptions2);
            Iterator it = allOptions2.iterator();
            while (it.hasNext()) {
                Option lo = (Option)it.next();
                OptionChooser loc = new OptionChooser(lo, allChooser);
                allChooser.add(loc);
            }
        }
        Iterator it = allChooser.iterator();
        while (it.hasNext()) {
            OptionChooser loc = (OptionChooser)it.next();
            if (loc.getOption().getAllInstances().size() > 1) {
                chooserBox.add(loc);
            }
        }
    }

    lavcGeneticMainDisplay(JFrame m, Map cm) {
        parent = m;
        parent.setVisible(false);
        codecMap = cm;
        
        java.awt.Dimension dim = new java.awt.Dimension(1024, 704);
        setMinimumSize(dim);
        setPreferredSize(dim);
        setLayout(new BoxLayout(this, BoxLayout.Y_AXIS));

        add(dcBox);
        dcBox.add(new JButton(new directoryChooserAction("Choose workdir")));
        dcBox.add(directoryChooserLabel);
        dcBox.add(new JButton(new sourceFileChooserAction("Choose source file")));
        dcBox.add(sourceFileChooserLabel);

        add(jtfExtraOpts);
        
        add(stBox);
        stBox.add(new JLabel("Initial number of run: "));
        JSlider stSlider = new JSlider(0, 300, nSimultaneousTests);
        stSlider.setSnapToTicks(true);
        stSlider.setMinorTickSpacing(3);
        stSlider.setMajorTickSpacing(15);
        stSlider.setPaintTicks(true);
        stSlider.setPaintLabels(true);
        stSlider.addChangeListener(new lavcSimultaneousTestsSliderListener());
        stBox.add(stSlider);

        add(rbBox);
        rbBox.add(new JLabel("Breeding Randomness: "));
        JSlider rbSlider = new JSlider(0, 100, nRandomBreeding);
        rbSlider.setSnapToTicks(false);
        rbSlider.setMinorTickSpacing(5);
        rbSlider.setMajorTickSpacing(25);
        rbSlider.setPaintTicks(true);
        rbSlider.setPaintLabels(true);
        rbSlider.addChangeListener(new lavcRandomBreedingSliderListener());
        rbBox.add(rbSlider);

        add(btBox);
        btBox.add(new JLabel("Bitrate: "));
        btSlider = new JSlider(MIN_BITRATE, MAX_BITRATE, bitrate);
        btSlider.setSnapToTicks(false);
        btSlider.setMinorTickSpacing(25);
        btSlider.setMajorTickSpacing(100);
        btSlider.setPaintTicks(true);
        btSlider.setPaintLabels(true);
        btSlider.addChangeListener(new lavcBitrateSliderListener());
        btTextField.setValue(new Integer(bitrate));
        btTextField.addPropertyChangeListener(new PropertyChangeListener() {
                public void propertyChange(PropertyChangeEvent e) {
                    if ("value".equals(e.getPropertyName())) {
                        Number value = (Number)e.getNewValue();
                        if (value != null) {
                            btSlider.setValue(value.intValue());
                        }
                    }
                }
            });
        java.awt.GridBagLayout btGBL = new java.awt.GridBagLayout();
        btBox.setLayout(btGBL);
        java.awt.GridBagConstraints c3 = new java.awt.GridBagConstraints();
        c3.weightx = 1.0;
        c3.fill = java.awt.GridBagConstraints.HORIZONTAL;
        c3.gridwidth = java.awt.GridBagConstraints.RELATIVE;
        btGBL.setConstraints(btSlider, c3);
        java.awt.GridBagConstraints c4 = new java.awt.GridBagConstraints();
        c4.weightx = 0.0;
        c4.gridwidth = java.awt.GridBagConstraints.REMAINDER;
        btGBL.setConstraints(btTextField, c4);
        btBox.add(btSlider);
        btBox.add(btTextField);

        add(bwAndfoiBox);

        bwAndfoiBox.add(bwBox);
        bwBox.add(new JLabel("Approx. Percentage: "));
        JSlider bwSlider = new JSlider(0, 200, nWithinPercentage);
        Dictionary dict = new Hashtable();
        for (int i = 0 ; i <= 200; i+=25) {
            dict.put(new Integer(i), new JLabel( "" + ((double)i/100.) ));
        }
        bwSlider.setLabelTable(dict);
        bwSlider.setSnapToTicks(true);
        bwSlider.setMinorTickSpacing(5);
        bwSlider.setMajorTickSpacing(25);
        bwSlider.setPaintTicks(true);
        bwSlider.setPaintLabels(true);
        bwSlider.addChangeListener(new lavcBestWithinSliderListener());
        bwBox.add(bwSlider);

        bwAndfoiBox.add(foiBox);
        foiBox.add(new JLabel("Float Options Interval: "));
        JSlider foiSlider = new JSlider(0, 6, 0);
        Dictionary foiDict = new Hashtable();
        foiDict.put(new Integer(0), new JLabel("Disabled"));
        for (int i = 1; i<= 6; i++) {
            foiDict.put(new Integer(i), new JLabel("" + lavcFloatIntervalCompute(i)));
        }
        foiSlider.setLabelTable(foiDict);
        foiSlider.setSnapToTicks(true);
        foiSlider.setMinorTickSpacing(1);
        foiSlider.setMajorTickSpacing(1);
        foiSlider.setPaintTicks(true);
        foiSlider.setPaintLabels(true);
        foiSlider.addChangeListener(new lavcFloatIntervalSliderListener());
        foiBox.add(foiSlider);

        Iterator it = codecMap.keySet().iterator();
        while (it.hasNext()) {
            String codec = (String)it.next();
            Box chooserBox = Box.createVerticalBox();
            addAllOptions(codec, chooserBox);
            chooserBox.setAlignmentX(java.awt.Component.LEFT_ALIGNMENT);
            chooserBox.setAlignmentY(java.awt.Component.TOP_ALIGNMENT);
            chooserBox.setName(codec);
            jtb.add(chooserBox);
            jtb.addChangeListener(new codecTabbedPaneChangeListener());
        }
        jtb.setSelectedIndex(0);
        add(mainPane);

        JCheckBox jcb = new JCheckBox(new AbstractAction("Use 2-Pass Encoding ") {
                public void actionPerformed(ActionEvent e) {
                    JCheckBox jcb = (JCheckBox)e.getSource();
                    twopass = jcb.isSelected();
                }
            });
        jcb.setSelected(true);

        add(bBox);
        bBox.add(jcb);
        bBox.add(new JButton(new AbstractAction("Explore selected codecs") {
                public void actionPerformed(ActionEvent e) {
                    cleanAndStart(new ArrayList(selectedCodec));
                }
            }));
        it = codecMap.keySet().iterator();
        while (it.hasNext()) {
            final String codec = (String)it.next();
            /*
            bBox.add(new JButton(new AbstractAction("Explore " + codec) {
                    public void actionPerformed(ActionEvent e) {
                        List codecs = new ArrayList();
                        codecs.add(codec);
                        cleanAndStart(codecs);
                    }
                }));
            */
            JCheckBox jcb2 = new JCheckBox(new selectCodecAction(codec));
            jcb2.setSelected(false);
            jcb2.doClick(); /* force to selected *with* call to the action */
            bBox.add(jcb2);
        }
        
        parent.setVisible(true);
    }

    void cleanAndStart(List codecs) {
        jtb.removeAll();
        remove(dcBox);
        remove(jtfExtraOpts);
        remove(stBox);
        remove(rbBox);
        remove(btBox);
        remove(bwAndfoiBox);
        remove(bBox);

        final lavcGeneticMain.allCodecsBarrier acb = new lavcGeneticMain.allCodecsBarrier(codecs.size());

        Iterator it = codecs.iterator();
        while (it.hasNext()) {
            final String startCodec = (String)it.next();
            final JPanel thisCodec = new JPanel();
            thisCodec.setLayout(new javax.swing.BoxLayout(thisCodec, javax.swing.BoxLayout.Y_AXIS));
            thisCodec.setName(startCodec);
            jtb.add(thisCodec);

            final Set pos = getProcessedOptionsSet(startCodec);
            final JTextArea jta = new JTextArea(pos.toString());
            jta.setLineWrap(true);
            jta.setWrapStyleWord(true);
            jta.setEditable(false);
            thisCodec.add(jta, 0);
            
            thisCodec.add(new JLabel("Codec: " + startCodec), 0);

            extraOpts = jtfExtraOpts.getText();
            
            final StatisticsDisplayPanel lasd = new StatisticsDisplayPanel(pos, parent);
            thisCodec.add(lasd);
            final Thread t = new Thread() {
                    public void run() {
                        lavcGeneticMain.doTheJob(startCodec,
                                                 pos,
                                                 lasd,
                                                 (lavcCodecData)codecMap.get(startCodec),
                                                 nSimultaneousTests,
                                                 nRandomBreeding,
                                                 bitrate,
                                                 twopass,
                                                 workdir,
                                                 (double)nWithinPercentage / 100.,
                                                 acb,
                                                 sourceFile,
                                                 extraOpts);
                    }
                };
            t.start();
        }
        add(new JButton(new AbstractAction("stop (after current step is done)") {
                public void actionPerformed(ActionEvent e) {
                    JButton jb = (JButton)e.getSource();
                    lavcGeneticMain.setDone();
                    jb.setEnabled(false);
                };
            }));
        parent.pack();
    }

    void setAverageParameterAllOptions(Object o) {
        Iterator it = chooserMap.keySet().iterator();
        while (it.hasNext()) {
            Set allChooser = (Set)chooserMap.get(it.next());
            Iterator it2 = allChooser.iterator();
            while (it2.hasNext()) {
                OptionChooser oc = (OptionChooser)it2.next();
                oc.getOption().setAverageParameter(o);
            }
        }
    }

    private Set getProcessedOptionsSet(String codec) {
        Set temp = new TreeSet();
        Iterator it = ((Set)chooserMap.get(codec)).iterator();
        while (it.hasNext()) {
            temp.add(((OptionChooser)it.next()).getProcessedOption());
        }
        return temp;
    }
}
