/* 
 * Copyright (C) 2004 Romain Dolbeau <romain@dolbeau.org>
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

package org.dolbeau.xsCOEj;

import java.util.*;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Attr;
import org.w3c.dom.NodeList;


/**
 * Create Options and OptionInstances from an XML Document.
 * @author Romain Dolbeau <romain@dolbeau.org>
 * @version  $Id: OptionFactoryXML.java,v 1.5 2005/05/20 15:01:24 dolbeau Exp $
 * @see Option
 * @see OptionInstance
 */
public class OptionFactoryXML implements java.io.Serializable {
    public static final String NAME_OPTION = "option";
    public static final String ATTR_OPTION_NAME = "name";
    public static final String ATTR_OPTION_PREFIX = "prefix";
    public static final String ATTR_OPTION_COMMENT = "comment";
    public static final String ATTR_OPTION_BEFORE = "before";
    public static final String ATTR_OPTION_AFTER = "after";
    public static final String ATTR_OPTION_MUSTCHOOSE = "must_choose";
    public static final String ATTR_OPTION_ISVIRTUAL = "virtual";
    public static final String ATTR_OPTION_DEFAULTCOMMAND = "commanddefault";
    public static final String ATTR_OPTION_DEFAULTSELECTED = "selecteddefault";
    public static final String ATTR_OPTION_DATATYPE = "datatype";
    public static final String ATTRVALUE_OPTION_DATATYPE_DOUBLE = "double";
    public static final String ATTRVALUE_OPTION_DATATYPE_INTEGER = "integer";
    public static final String ATTR_OPTION_INTEGER_PRECISION = "integer_precision";

    public static final String NAME_OPTIONINSTANCE = "optioninstance";
    public static final String ATTR_OPTIONINSTANCE_NAME = "name";
    public static final String NAME_INCOMPATIBLE = "incompatible";
    public static final String NAME_REQUIRE = "require";
    public static final String ATTR_INCOMPATIBLE_OPTIONNAME = "optionname";
    public static final String ATTR_INCOMPATIBLE_OPTIONINSTANCENAME = "optioninstancename";

    public static final String NAME_LIST = "list";
    public static final String NAME_INSTANCE = "instance";
    public static final String ATTR_INSTANCE_OPTIONNAME = ATTR_INCOMPATIBLE_OPTIONNAME;
    public static final String ATTR_INSTANCE_OPTIONINSTANCENAME = ATTR_INCOMPATIBLE_OPTIONINSTANCENAME;

    private final Map nameToOption = new HashMap();

    private Set all = null;
    private Set allLists = null;
    
    public OptionFactoryXML(Document doc) {
        all = buildAllOptions(doc);
        allLists = buildAllLists(doc, all);
    }

    public Set getAllOptions() {
        return new TreeSet(all);
    }

    public Set getDefaultLists() {
        return new TreeSet(allLists);
    }


    private OptionFromXML getOptionByName(String n) {
        return (OptionFromXML)nameToOption.get(n);
    }
    
    protected class OptionFromXML extends Option {
        private Map nameToOptionInstance = new HashMap();

        private void setMustChoose(boolean mc) {
            mChoose = mc;
        }
        private void setIsVirtual(boolean mc) {
            virtualOpt = mc;
        }

        public Option getOneInstanceOption(OptionInstance oi) {
            if ((before.size() > 0) || (after.size() > 0))
                System.err.println("WARNING: calling getOneInstanceOption() on ordered Option is NOT SUPPORTED");
            return new OptionFromXML(oi);
        }
        
        private OptionFromXML(OptionInstance oi) {
            prefix = ((OptionFromXML)oi.getParentOption()).prefix;
            desc = ((OptionFromXML)oi.getParentOption()).desc;
            comment = ((OptionFromXML)oi.getParentOption()).comment;
            instances.add(oi);
        }

        private OptionInstanceFromXML getOptionInstanceByName(String n) {
            return (OptionInstanceFromXML)nameToOptionInstance.get(n);
        }
        
        protected class OptionInstanceFromXML extends OptionInstanceIncompatibleList {
            protected final String desc;
            
            public String getOptionString() {
                if ((prefix != null) && (desc != null))
                    return prefix + desc;
                else
                    return desc;
            }
            
            protected OptionInstanceFromXML(Option o, String d) {
                super(o);
                desc = d;
                
                nameToOptionInstance.put(d == null ? "" : d, this);
            }
            
            private OptionInstanceFromXML(Option o, Element e) {
                super(o);
                String temp = e.getAttribute(ATTR_OPTIONINSTANCE_NAME);

                if (temp.equals(""))
                    desc = null;
                else
                    desc = temp;

                nameToOptionInstance.put(temp, this); // temp not desc, so that the mapping uses "" and not (null)
            }

            private void setIncompatibleFromXML(Element e) {
                NodeList l = e.getElementsByTagName(NAME_INCOMPATIBLE);
                final int n = l.getLength();

                for (int i = 0 ; i < n ; i++) {
                    Element c = (Element)l.item(i);
                    String on = c.getAttribute(ATTR_INCOMPATIBLE_OPTIONNAME);
                    OptionFromXML o = getOptionByName(on);
                    Attr attr =  c.getAttributeNode(ATTR_INCOMPATIBLE_OPTIONINSTANCENAME);
                    if (attr != null) {
                        String oin = attr.getValue();
                        OptionInstanceFromXML oi = o.getOptionInstanceByName(oin);
                        this.addIncompatible(oi);
                        oi.addIncompatible(this);
                    } else { // incompatible with everything but ""
                        List l2 = o.getAllInstances();
                        Iterator it2 = l2.iterator();
                        while (it2.hasNext()) {
                            OptionInstanceIncompatibleList oi = (OptionInstanceIncompatibleList)it2.next();
                            if (oi.getOptionString() != null) {
                                this.addIncompatible(oi);
                                oi.addIncompatible(this);
                            }
                        }
                    }
                }

                
                NodeList l2 = e.getElementsByTagName(NAME_REQUIRE);
                final int n2 = l2.getLength();
                for (int i = 0 ; i < n2 ; i++) {
                    Element c = (Element)l2.item(i);
                    String on = c.getAttribute(ATTR_INCOMPATIBLE_OPTIONNAME);
                    OptionFromXML o = getOptionByName(on);
                    Attr attr =  c.getAttributeNode(ATTR_INCOMPATIBLE_OPTIONINSTANCENAME);
                    if (attr != null) {
                        String oin = attr.getValue();
                        OptionInstanceIncompatibleList oi = o.getOptionInstanceByName(oin);
                        List list = o.getAllInstancesBut(oi);
                        Iterator it = list.iterator();
                        while (it.hasNext()) {
                            OptionInstanceIncompatibleList oi2 = (OptionInstanceIncompatibleList)it.next();
                            this.addIncompatible(oi2);
                            oi2.addIncompatible(this);
                        }
                    } else {
                        // TODO: handle problem
                    }
                }
            }
        }

        protected final String desc;
        protected final String comment;
        protected final String prefix;
        protected final List instances = Collections.synchronizedList(new ArrayList());
        
        public String getComment() {
            return comment;
        }

        public String getDescription() {
            return desc;
        }

        public List getAllInstances() {
            List temp = null;
            synchronized (instances) {
                temp = new ArrayList(instances);
            }
            return temp;
        }

        private OptionFromXML(Element e) {
            Attr attr = e.getAttributeNode(ATTR_OPTION_PREFIX);
            if (attr != null) {
                prefix = attr.getValue();
                attr = null;
            } else {
                prefix = null;
            }
            attr = e.getAttributeNode(ATTR_OPTION_COMMENT);
            if (attr != null) {
                comment = attr.getValue();
                attr = null;
            } else {
                comment = null;
            }
            attr = e.getAttributeNode(ATTR_OPTION_MUSTCHOOSE);
            if (attr != null) {
                String mc = attr.getValue();
                if (mc.equals("true"))
                    setMustChoose(true);
            }
            attr = e.getAttributeNode(ATTR_OPTION_ISVIRTUAL);
            if (attr != null) {
                String mc = attr.getValue();
                if (mc.equals("true"))
                    setIsVirtual(true);
            }
            

            desc = e.getAttribute(ATTR_OPTION_NAME);

            NodeList l = e.getElementsByTagName(NAME_OPTIONINSTANCE);
            final int n = l.getLength();
            for (int i = 0 ; i < n ; i++) {
                Element c = (Element)l.item(i);
                instances.add(new OptionInstanceFromXML(this, c));
            }
            nameToOption.put(desc, this);

            attr = e.getAttributeNode(ATTR_OPTION_DEFAULTCOMMAND);
            if (attr != null)
                setDefaultCommandIndex(getOptionInstanceByName(attr.getValue()));
            attr = e.getAttributeNode(ATTR_OPTION_DEFAULTSELECTED);
            if (attr != null)
                setDefaultSelectedIndex(getOptionInstanceByName(attr.getValue()));
        }

        private void setIncompatibleFromXML(Element e) {
            NodeList l = e.getElementsByTagName(NAME_OPTIONINSTANCE);
            final int n = l.getLength();
            for (int i = 0 ; i < n ; i++) {
                Element c = (Element)l.item(i);
                OptionInstanceFromXML oi = getOptionInstanceByName(c.getAttribute(ATTR_OPTIONINSTANCE_NAME));
                oi.setIncompatibleFromXML(c);
            }
        }
    }

    private class IntegerOptionFromXML extends OptionFromXML {
        private int interval = 0;

        private IntegerOptionFromXML(Element e) {
            super(e);
        }

        public boolean canUseAverage() {
            return true;
        }

        public void setAverageParameter(Object o) {
            if (o instanceof Integer)  {
                interval = ((Integer)o).intValue();
            } else {
                // nothing, it's not for us
            }
        }

        private OptionInstanceFromXML addInstanceIfNew(int f) {
            OptionInstanceFromXML temp = null;
            
            synchronized (instances) {
                Iterator it = instances.iterator();
                while (it.hasNext() && (temp == null)) {
                    OptionInstanceFromXML loif = (OptionInstanceFromXML)it.next();
                    int v = Integer.parseInt(loif.desc);
                    if (v == f) {
                        temp = loif;
                    }
                }
                if (temp == null) {
                    temp = new OptionInstanceFromXML(this, "" + f);
                    instances.add(temp);
                }
            }
            return temp;
        }

        public OptionInstance breed(OptionInstance loi1, OptionInstance loi2,
                                    int oddsR, Random random) {
            if ((!useAverage) ||
                (interval == 0))
                return super.breed(loi1, loi2, oddsR, random);
            
            if (random.nextInt(100) < oddsR) {
                List all = getAllInstances();
                return (OptionInstance)all.get(random.nextInt(all.size()));
            }

System.err.print("OK: breeding " + desc + " between " + loi1.getOptionString() + " and " + loi2.getOptionString());
            
            OptionInstanceFromXML loif1 = (OptionInstanceFromXML)loi1;
            OptionInstanceFromXML loif2 = (OptionInstanceFromXML)loi2;
            
            int v1 = Integer.parseInt(loif1.desc);
            int v2 = Integer.parseInt(loif2.desc);
            
            int vm = (v1 + v2 + random.nextInt(2)) / 2;

            if ((vm % interval) != 0) {
                OptionInstanceFromXML temp = null;
                synchronized (instances) {
                    Iterator it = instances.iterator();
                    while (it.hasNext() && (temp == null)) {
                        OptionInstanceFromXML loif = (OptionInstanceFromXML)it.next();
                        int v = Integer.parseInt(loif.desc);
                        if (v == vm) {
                            temp = loif;
                        }
                    }
                }
                if (temp == null) {
                    int vd1 = (vm - (vm % interval));
                    int vd2 = vd1 + interval;
                    if ((vm - vd1) == (vd2 - vm)) {
                        vm = vd1 + (interval * random.nextInt(2));
                    } else if ((vm - vd1) < (vd2 - vm)) {
                        vm = vd1;
                    } else if ((vm - vd1) > (vd2 - vm)) {
                        vm = vd2;
                    }
                }
            }

System.err.println(" gave " + vm);
            
            return addInstanceIfNew((int)vm);
        }
    }


    private class DoubleOptionFromXML extends OptionFromXML {
        private double interval = 0.;

        private DoubleOptionFromXML(Element e) {
            super(e);
        }

        public boolean canUseAverage() {
            return true;
        }

        public void setAverageParameter(Object o) {
            if (o instanceof Double)  {
                interval = ((Double)o).doubleValue();
            } else {
                // nothing, it's not for us
            }
        }

        private final static double epsilon = 0.000001;
        private OptionInstanceFromXML addInstanceIfNew(double f) {
            OptionInstanceFromXML temp = null;
            
            synchronized (instances) {
                Iterator it = instances.iterator();
                while (it.hasNext() && (temp == null)) {
                    OptionInstanceFromXML loif = (OptionInstanceFromXML)it.next();
                    double v = Double.parseDouble(loif.desc);
                    if (Math.abs(v - f) < epsilon) {
                        temp = loif;
                    }
                }
                if (temp == null) {
                    temp = new OptionInstanceFromXML(this, "" + f);
                    instances.add(temp);
                }
            }
            return temp;
        }

        public OptionInstance breed(OptionInstance loi1, OptionInstance loi2,
                                    int oddsR, Random random) {

            if ((!useAverage) ||
                (interval == (double)0.))
                return super.breed(loi1, loi2, oddsR, random);
            
            if (random.nextInt(100) < oddsR) {
                List all = getAllInstances();
                return (OptionInstance)all.get(random.nextInt(all.size()));
            }
System.err.println("OK: breeding " + desc + " between " + loi1.getOptionString() + " and " + loi2.getOptionString());
            
            OptionInstanceFromXML loif1 = (OptionInstanceFromXML)loi1;
            OptionInstanceFromXML loif2 = (OptionInstanceFromXML)loi2;
            
            double v1 = Double.parseDouble(loif1.desc);
            double v2 = Double.parseDouble(loif2.desc);
            
            double vm = (v1 + v2) / 2.;
            double newvm;
            
            double vMin = (Math.floor(vm / interval)) * interval;
            double vMax = (Math.ceil(vm / interval)) * interval;
            
            if ((vm - vMin) <= (vMax - vm)) {
                newvm = vMin;
            } else {
                newvm = vMax;
            }
System.err.println(" gave " + newvm);

            return addInstanceIfNew(newvm);
        }
    }
    
    private Set buildAllOptions(Document doc) {
        Set allOptions = new HashSet();

        Element root = doc.getDocumentElement();

        NodeList l = root.getElementsByTagName(NAME_OPTION);
        
        final int n = l.getLength();
        for (int i = 0 ; i < n ; i++) {
            Element e = (Element)l.item(i);
            OptionFromXML o = null;

            Attr attr = e.getAttributeNode(ATTR_OPTION_DATATYPE);
            if (attr != null) {
                if (attr.getValue().equals(ATTRVALUE_OPTION_DATATYPE_DOUBLE)) {
                    o = new DoubleOptionFromXML(e);
                } else {
                    if (attr.getValue().equals(ATTRVALUE_OPTION_DATATYPE_INTEGER)) {
                        o = new IntegerOptionFromXML(e);
                        ((IntegerOptionFromXML)o).setAverageParameter(new Integer(1));
                        Attr attr36 = e.getAttributeNode(ATTR_OPTION_INTEGER_PRECISION);
                        if (attr36 != null) {
                            int val = Integer.parseInt(attr36.getValue());
                            ((IntegerOptionFromXML)o).setAverageParameter(new Integer(val));
                        }
                    } else {
                        // TODO: handle new cases
                    }
                }
            } else {
                o = new OptionFromXML(e);
            }
            allOptions.add(o);
        }
        for (int i = 0 ; i < n ; i++) {
            Element e = (Element)l.item(i);
            OptionFromXML o = getOptionByName(e.getAttribute(ATTR_OPTION_NAME));
            if (o == null) {
                System.err.println("OUPS (option not found): " + e.getAttribute(ATTR_OPTION_NAME));
            }
            o.setIncompatibleFromXML(e);

            Attr attr;
            attr = e.getAttributeNode(ATTR_OPTION_BEFORE);
            if (attr != null) {
                String beforeS = attr.getValue();
                Option before = getOptionByName(beforeS);
                o.addBefore(before);
            }
            attr = e.getAttributeNode(ATTR_OPTION_AFTER);
            if (attr != null) {
                String afterS = attr.getValue();
                Option after = getOptionByName(afterS);
                if (after == null) {
                    System.err.println("OUPS (after): " + e.getAttribute(ATTR_OPTION_NAME));
                }
                after.addBefore(o);
            }
        }
        /*
        for (int i = 0 ; i < n ; i++) {
            Element e = (Element)l.item(i);
            OptionFromXML o = getOptionByName(e.getAttribute(ATTR_OPTION_NAME));
            System.err.println("" + o + " HAS BEFORE " + o.getBefore());
        }
        */
        return allOptions;
    }

    private OptionInstanceList listFromXML(Element el, Set allOptions) {
        OptionInstanceList oil = new OptionInstanceList();

        NodeList l = el.getElementsByTagName(NAME_INSTANCE);

        final int n = l.getLength();
        for (int i = 0 ; i < n ; i++) {
            Element e = (Element)l.item(i);
            String on = e.getAttribute(ATTR_INSTANCE_OPTIONNAME);
            OptionFromXML o = getOptionByName(on);
            if (o == null) {
                System.err.println("Uknown option: " + on);
            } else {
                String oin = e.getAttribute(ATTR_INSTANCE_OPTIONINSTANCENAME);
                OptionFromXML.OptionInstanceFromXML oi = o.getOptionInstanceByName(oin);
                if (oi != null)
                    oil.add(oi);
                else
                    System.err.println("Uknown instance: " + oin + " for option " + on);
            }
        }

        Iterator it = allOptions.iterator();
        while (it.hasNext()) {
            Option o = (Option)it.next();
            OptionInstance oi = oil.getOptionInstance(o);
            if (oi == null)
                oil.add(o.getDefaultInstance());
        }

        return oil;
    }
    
    private Set buildAllLists(Document doc, Set allOptions) {
        Set allLists = new HashSet();

        Element root = doc.getDocumentElement();

        NodeList l = root.getElementsByTagName(NAME_LIST);

        final int n = l.getLength();
        for (int i = 0 ; i < n ; i++) {
            Element e = (Element)l.item(i);
            OptionInstanceList oil = listFromXML(e, allOptions);

            allLists.add(oil);
        }

        return allLists;
    }
}
