/* 
 * Copyright (C) 2004 Romain Dolbeau <romain@dolbeau.org>
 *
 *  This file is subject to the terms and conditions of the GNU General Public
 *  License. See the file COPYING in the main directory of this archive for
 *  more details.
 */

package org.dolbeau.xsCOEj;

import java.util.*;

/**
 * The Class representing the pool of results and the associated options instances.
 * @author Romain Dolbeau <romain@dolbeau.org>
 * @version  $Id: ResultsPool.java,v 1.1.1.1 2004/12/06 13:52:41 dolbeau Exp $
 * @see Results
 * @see OptionInstanceList
 */
public class ResultsPool {
    private Map allResults = null;
    private int invalidCount = 0;

    /** Build an empty ResultsPool */
    public ResultsPool() {
        allResults = Collections.synchronizedMap(new HashMap());
    }

    /** Build a ResultsPool with the same entries as another ResultsPool
     * @param lrp The ResultsPool to duplicate
     */
    public ResultsPool(ResultsPool lrp) {
        synchronized (lrp.allResults) {
            allResults = Collections.synchronizedMap(new HashMap(lrp.allResults));
        }
    }

    /** Check if a particular OptionInstanceList already has a result in this ResultsPool.
     * @param opt The OptionInstanceList to check.
     * @return If the OptionInstanceList has a valid result in this ResultsPool
     */
    public boolean hasResult(OptionInstanceList opt) {
        return getResult(opt) != null;
    }

    /** Get the Results for a particular OptionInstanceList in this ResultsPool.
     * @param opt The OptionInstanceList to check.
     * @return The Results for the OptionInstanceList in this ResultsPool.
     */
    public Results getResult(OptionInstanceList opt) {
        Results result = null;
        synchronized (allResults) {
            if (allResults.containsKey(opt)) {
                result = (Results)allResults.get(opt);
            }
        }
        return result;
    }
    
    public void addResult(OptionInstanceList opt, Results r) {
        if (hasResult(opt)) {
            Results r2 = getResult(opt);
            System.err.println("ERROR !!!! *not* adding an already present result for " + opt + " new result: " + r + " old result: " + r2);
        } else {
            synchronized (allResults) {
                allResults.put(opt, r);
                if (!r.isValid())
                    invalidCount++;
            }
        }
    }

    private void addResultIgnoreDup(OptionInstanceList opt, Results r) {
        if (!hasResult(opt)) {
            synchronized (allResults) {
                allResults.put(opt, r);
                if (!r.isValid())
                    invalidCount++;
            }
        }
    }

    public void mergeResults(ResultsPool lrp) {
        synchronized (lrp.allResults) {
            Iterator it = lrp.allResults.keySet().iterator();
            while (it.hasNext()) {
                OptionInstanceList opt = (OptionInstanceList)it.next();
                Results result = (Results)lrp.allResults.get(opt);
                addResultIgnoreDup(opt, result);
            }
        }
    }

    public int getInvalidCount() {
        int temp = 0;
        synchronized (allResults) {
            temp = invalidCount;
        }
        return temp;
    }
    

    public int numberOfResults() {
        int n = -1;

        synchronized (allResults) {
            n = allResults.size();
        }

        return n;
    }

    public Map getAllResults() {
        Map copy;

        synchronized (allResults) {
            copy = new HashMap(allResults);
        }

        return copy;
    }

    public OptionInstanceList getBestResultKey() {
        Results bestResult = null;
        OptionInstanceList best = null;

        synchronized (allResults) {
            if (allResults.size() > 0) {
                Iterator it = allResults.keySet().iterator();
                best = (OptionInstanceList)it.next();
                bestResult = (Results)allResults.get(best);
                while (it.hasNext()) {
                    OptionInstanceList temp = (OptionInstanceList)it.next();
                    Results tempResult = (Results)allResults.get(temp);

                    if (tempResult.compareTo(bestResult) > 0) {
                        best = temp;
                        bestResult = tempResult;
                    }
                }
            }
        }
        return best;
    }

    public OptionInstanceList getWorstResultKey() {
        Results worstResult = null;
        OptionInstanceList worst = null;

        synchronized (allResults) {
            if (allResults.size() > 0) {
                Iterator it = allResults.keySet().iterator();
                do {
                    worst = (OptionInstanceList)it.next();
                    worstResult = (Results)allResults.get(worst);
                } while ((worst != null) && (it.hasNext()) && (!worstResult.isValid()));
                if (worst != null) {
                    while (it.hasNext()) {
                        OptionInstanceList temp = (OptionInstanceList)it.next();
                        Results tempResult = (Results)allResults.get(temp);
                        
                        if ((tempResult.compareTo(worstResult) < 0) &&
                            (tempResult.isValid())) {
                            worst = temp;
                            worstResult = tempResult;
                        }
                    }
                }
            }
        }
        return worst;
    }

    public Results getBestResult() {
        OptionInstanceList best = getBestResultKey();
        Results bestResult = null;
        synchronized (allResults) {
            bestResult = (Results)allResults.get(best);
        }
        return bestResult;
    }

    private void removeResult(OptionInstanceList l) {
        synchronized (allResults) {
            Results result = (Results)allResults.get(l);
            allResults.remove(l);
            if (!result.isValid())
                invalidCount--;
        }
    }

    private void removeAllInvalid() {
        synchronized (allResults) {
            Iterator it = allResults.keySet().iterator();
            while (it.hasNext()) {
                List l = (List)it.next();
                Results r = (Results)allResults.get(l);
                if (!r.isValid()) {
                    it.remove(); // don't use allResults.remove(l) to avoid destroying iterator
                }
            }
        }
    }

    public OptionInstanceList getFastestKey() {
        Results bestResult = null;
        OptionInstanceList best = null;

        synchronized (allResults) {
            if (allResults.size() > 0) {
                Iterator it = allResults.keySet().iterator();
                do {
                    best = (OptionInstanceList)it.next();
                    bestResult = (Results)allResults.get(best);
                } while ((best != null) && (it.hasNext()) && (!bestResult.isValid()));
                if (best != null) {
                    while (it.hasNext()) {
                        OptionInstanceList temp = (OptionInstanceList)it.next();
                        Results tempResult = (Results)allResults.get(temp);
                        
                        if ((tempResult.getSeconds() < bestResult.getSeconds()) &&
                            (tempResult.getSeconds() > 0)) {
                            best = temp;
                            bestResult = tempResult;
                        }
                    }
                }
            }
        }
        return best;
    }

    public List getNBestKeys(int n) {
        ResultsPool copy = new ResultsPool(this);
        List temp = new ArrayList();
        for (int i = 0 ; (i < n) && (copy.numberOfResults() > 0) ; i++) {
            OptionInstanceList l = copy.getBestResultKey();
            copy.removeResult(l);
            temp.add(l);
        }
        return temp;
    } 

    public ResultsPool getNBestResults(int n) {
        ResultsPool copy = new ResultsPool(this);
        ResultsPool temp = new ResultsPool();
        for (int i = 0 ; (i < n) && (copy.numberOfResults() > 0) ; i++) {
            OptionInstanceList l = copy.getBestResultKey();
            Results r = copy.getResult(l);
            copy.removeResult(l);
            temp.addResult(l, r);
        }
        return temp;
    }

    public List getNWorstKeys(int n) {
        ResultsPool copy = new ResultsPool(this);
        List temp = new ArrayList();
        for (int i = 0 ; (i < n) && (copy.numberOfResults() > 0) ; i++) {
            OptionInstanceList l = copy.getWorstResultKey();
            if (l != null) {
                copy.removeResult(l);
                temp.add(l);
            }
        }
        return temp;
    }

    public ResultsPool getAllButHalfTheWorstN(int n, Random random) {
        ResultsPool copy = new ResultsPool(this);
        copy.removeAllInvalid();
        List worst = copy.getNWorstKeys(n);
        for (int i = 0; (i < n / 2) && (worst.size() > 0) ; i ++) {
            Object temp = worst.get(random.nextInt(worst.size()));
            worst.remove(temp);
        }
        Iterator it = worst.iterator();
        while (it.hasNext()) {
            copy.removeResult((OptionInstanceList)it.next());
        }
        return copy;
    }

    public OptionInstanceList fasterWithinPercentBestKey(double p) {
        ResultsPool copy = new ResultsPool(this);
        ResultsPool best = new ResultsPool();
        OptionInstanceList l = copy.getBestResultKey();
        Results b = getResult(l);
        boolean done = false;

        while ((copy.numberOfResults() > 0) && !done) {
            OptionInstanceList l2 = copy.getBestResultKey();
            Results b2 = copy.getResult(l2);
            copy.removeResult(l2);
            if (b2.isWithinPercent(b, p)) {
                best.addResult(l2, b2);
            } else {
                done = true;
            }
        }
        return best.getFastestKey();
    }

    public Set getSetOfKeys() {
        Set theSet;
        synchronized (allResults) {
            theSet = allResults.keySet();
        }
        return new TreeSet(theSet);
    }

    public String toString() {
        return "ResultsPool: " + allResults;
    }

    public void dumpAllResults(java.io.Writer w) throws java.io.IOException {
        synchronized (allResults) {
            Set entries = allResults.entrySet();
            List entriesList = new ArrayList(entries);
            Collections.sort(entriesList, new Comparator() {
                    public int compare(Object o1, Object o2) {
                        Map.Entry e1 = (Map.Entry)o1;
                        Map.Entry e2 = (Map.Entry)o2;
                        Results r1 = (Results)e1.getValue();
                        Results r2 = (Results)e2.getValue();
                        if (r1 != null)
                            return r1.compareTo(r2);
                        if (r2 != null)
                            return r2.compareTo(r1);
                        return 0;
                    }
                    public boolean equals(Object obj) {
                        return false;
                    }
                });
            Iterator it = entriesList.iterator();
            while (it.hasNext()) {
                Map.Entry e = (Map.Entry)it.next();
                w.write(e.getKey() + " := " + e.getValue() + "\n");
            }
        }
    }

    public Map getStatisticsMap() {
        Map map = new HashMap();
        
        synchronized (allResults) {
            Iterator it = allResults.keySet().iterator();
            while (it.hasNext()) {
                OptionInstanceList l = (OptionInstanceList)it.next();
                Iterator it2 = l.iterator();
                while (it2.hasNext()) {
                    OptionInstance loi = (OptionInstance)it2.next();
                    if (map.containsKey(loi)) {
                        Integer old = (Integer)map.get(loi);
                        map.put(loi, new Integer(old.intValue() + 1));
                    } else {
                        map.put(loi, new Integer(1));
                    }
                }
            }
        }

        return map;
    }

    public String optionStatistics(Set allOptions) {
        Map map = getStatisticsMap();
        double n = (double)numberOfResults();
        String temp = "";
        Iterator it = allOptions.iterator();
        while (it.hasNext()) {
            Option opt = (Option)it.next();
            temp = temp + "\t" + opt.getDescription() + ": ";
            Iterator it2 = opt.getAllInstances().iterator();
            while (it2.hasNext()) {
                OptionInstance loi = (OptionInstance)it2.next();
                Integer a;
                if (map.containsKey(loi))
                    a = (Integer)map.get(loi);
                else
                    a = new Integer(0);
                temp = temp + "\"" + loi.toString() + "\" = " + a + " (" + (a.doubleValue() * 100. / n) + " %) ";
            }
            temp = temp + "\n";
        }
        return temp;
    }
    
    public void dumpEachOptionBenefit(java.io.Writer w,
                                      OptionInstanceList l,
                                      Set allSelectedOptions) throws java.io.IOException {
        Results rl = (Results)getResult(l);
        
        w.write(l + " := " + rl + "\n");
        
        Iterator it = allSelectedOptions.iterator();
        while (it.hasNext()) {
            Option opt = (Option)it.next();
            List allInstances = opt.getAllInstances();
            if (allInstances.size() > 1) {
                OptionInstance mine = l.getOptionInstance(opt);
                w.write("\tOption: " + opt.getDescription() + " (has: " + mine + ")\n");
                Iterator it2 = allInstances.iterator();
                while (it2.hasNext()) {
                    OptionInstance loi = (OptionInstance)it2.next();
                    if (loi.compareTo(mine) != 0) {
                        OptionInstanceList l2 = new OptionInstanceList(l);
                        l2.remove(mine);
                        l2.add(loi);
                        Collections.sort(l2);
                        Results rl2 = (Results)getResult(l2);
                        if (rl2 != null) {
                            double diff =(rl.getMetricAsADouble() - rl2.getMetricAsADouble());
                            if (!rl.isBiggerBetter())
                                diff = -1. * diff;
                            double base = rl.isBiggerBetter() ? rl2.getMetricAsADouble() : rl.getMetricAsADouble();
                            w.write("\t\t[" + loi + "] := " + rl2 +
                                    " (Metric gain: " + 
                                    diff +
                                    ", " +
                                    ((diff / base) * 100.) + " %)" +
                                    " (Time loss: " +
                                    (rl.getSeconds() - rl2.getSeconds()) +
                                    ", " +
                                    (((rl.getSeconds() - rl2.getSeconds()) / (rl2.getSeconds())) * 100.) + " %)" +
                                    "\n");
                        } else {
                            w.write("\t\t[" + loi + "] not tested\n");
                        }
                    }
                }
            }
        }
    }
}
